"""add_auto_terminate_col

Revision ID: 857bdec5abda
Revises: 210c4d9be768
Create Date: 2023-05-23 12:18:02.046409

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "857bdec5abda"
down_revision = "210c4d9be768"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "agents",
        sa.Column(
            "auto_terminate_abusing_kernel",
            sa.Boolean(),
            server_default=sa.text("false"),
            nullable=False,
        ),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("agents", "auto_terminate_abusing_kernel")
    # ### end Alembic commands ###
